package statistics

import (
	"context"
	"git.noahlan.cn/northlan/ntools-go/gorm-zero/gormx"
	"github.com/pkg/errors"
	"github.com/shopspring/decimal"
	"github.com/zeromicro/go-zero/core/logx"
	"gorm.io/gorm"
	"live-service/app/user_center/model"
	"live-service/app/user_center/rpc/internal/common/grade"
	"live-service/app/user_center/rpc/internal/svc"
	"live-service/app/user_center/rpc/pb"
	"sort"
)

type StatPvpReportLogic struct {
	ctx    context.Context
	svcCtx *svc.ServiceContext
	logx.Logger
}

func NewStatPvpReportLogic(ctx context.Context, svcCtx *svc.ServiceContext) *StatPvpReportLogic {
	return &StatPvpReportLogic{
		ctx:    ctx,
		svcCtx: svcCtx,
		Logger: logx.WithContext(ctx),
	}
}

type (
	sumType struct {
		damage     int64
		deDamage   int64
		killPlayer int64
		killUnit   int64
		deKillUnit int64
	}

	statPvPReportItem struct {
		respItem     *pb.StatPvPReportResp_Item
		uid          int64
		uname        string
		damage       int64
		deDamage     int64
		killUnit     int64
		deKillUnit   int64
		firstBlood   bool
		deFirstBlood bool
		killPlayer   int64
		deKillPlayer bool
		isGeneral    bool
	}
)

func (l *StatPvpReportLogic) StatPvpReport(in *pb.StatPvPReportReq) (*pb.StatPvPReportResp, error) {
	var err error
	resp := &pb.StatPvPReportResp{
		WinCamp:  in.WinCamp,
		BattleId: in.BattleId,
	}
	// 1. 计算评分
	winSum, winItems := l.calcSum(in.WinItems)
	lostSum, lostItems := l.calcSum(in.LostItems)
	l.calcScoreResponse(winSum, winItems)
	l.calcScoreResponse(lostSum, lostItems)

	// 2. 评分排序
	sort.Slice(winItems, func(i, j int) bool {
		return winItems[i].respItem.Score > winItems[j].respItem.Score
	})
	sort.Slice(lostItems, func(i, j int) bool {
		return lostItems[i].respItem.Score > lostItems[j].respItem.Score
	})

	// 3. 添加position + 各项记录(可忽略错误)
	err = l.reports(true, winItems)
	if err != nil {
		l.Logger.Errorf("获胜PvP记录失败, err:%+v", err)
	}
	err = l.reports(false, lostItems)
	if err != nil {
		l.Logger.Errorf("战败PvP记录失败, err:%+v", err)
	}

	// 4. 段位
	winGradeItems := make([]grade.ReportReq, 0, len(winItems))
	lostGradeItems := make([]grade.ReportReq, 0, len(lostItems))
	for _, item := range winItems {
		winGradeItems = append(winGradeItems, grade.ReportReq{
			UserId:   item.uid,
			Username: item.uname,
			Position: item.respItem.Position,
			Score:    item.respItem.Score,
		})
	}
	for _, item := range lostItems {
		lostGradeItems = append(lostGradeItems, grade.ReportReq{
			UserId:   item.uid,
			Username: item.uname,
			Position: item.respItem.Position,
			Score:    item.respItem.Score,
		})
	}
	if gradeReported, err := l.svcCtx.GradeManager.Report(l.ctx, winGradeItems, lostGradeItems); err != nil {
		l.Logger.Errorf("计算段位时发生错误", err)
	} else {
		for _, reportResp := range gradeReported {
			for _, item := range winItems {
				if item.uid == reportResp.UserId {
					item.respItem.Grade = reportResp.Grade
					item.respItem.GradeResult = reportResp.Result
					item.respItem.GradeReason = reportResp.Reason
				}
			}
			for _, item := range lostItems {
				if item.uid == reportResp.UserId {
					item.respItem.Grade = reportResp.Grade
					item.respItem.GradeResult = reportResp.Result
					item.respItem.GradeReason = reportResp.Reason
				}
			}
		}
	}

	// resp
	resp.WinItems = make([]*pb.StatPvPReportResp_Item, 0, len(winItems))
	resp.LostItems = make([]*pb.StatPvPReportResp_Item, 0, len(lostItems))
	for _, item := range winItems {
		resp.WinItems = append(resp.WinItems, item.respItem)
	}
	for _, item := range lostItems {
		resp.LostItems = append(resp.LostItems, item.respItem)
	}
	return resp, nil
}

func (l *StatPvpReportLogic) calcSum(items []*pb.StatPvPReportReq_Item) (sumType, []*statPvPReportItem) {
	resp := make([]*statPvPReportItem, 0, len(items))
	var sum sumType
	for _, item := range items {
		sum.damage += item.Damage
		sum.deDamage += item.DeDamage
		sum.killPlayer += item.KillPlayer
		sum.killUnit += item.KillUnit
		sum.deKillUnit += item.DeKillUnit
		resp = append(resp, &statPvPReportItem{
			uid:          item.Uid,
			uname:        item.Uname,
			damage:       item.Damage,
			deDamage:     item.DeDamage,
			killUnit:     item.KillUnit,
			deKillUnit:   item.DeKillUnit,
			firstBlood:   item.FirstBlood,
			deFirstBlood: item.DeFirstBlood,
			killPlayer:   item.KillPlayer,
			deKillPlayer: item.DeKillPlayer,
			isGeneral:    item.IsGeneral,
		})
	}
	return sum, resp
}

func (l *StatPvpReportLogic) calcScoreResponse(sum sumType, items []*statPvPReportItem) {
	for _, item := range items {
		itemResp := &pb.StatPvPReportResp_Item{
			Uid:   item.uid,
			Uname: item.uname,
			Score: 0,
			Grade: &pb.Grade{},
		}
		var tmpScore float64
		if sum.damage != 0 {
			tmp, _ := decimal.NewFromInt(item.damage).
				Div(decimal.NewFromInt(sum.damage)).
				Mul(decimal.NewFromFloat32(0.45)).
				Mul(decimal.NewFromInt(100)).Float64()
			tmpScore += tmp
		}
		if sum.damage != 0 {
			tmp, _ := decimal.NewFromInt(item.deDamage).
				Div(decimal.NewFromInt(sum.deDamage)).
				Mul(decimal.NewFromFloat32(0.15)).
				Mul(decimal.NewFromInt(100)).Float64()
			tmpScore += tmp
		}
		if sum.killPlayer != 0 {
			tmp, _ := decimal.NewFromInt(item.killPlayer).
				Div(decimal.NewFromInt(sum.killPlayer)).
				Mul(decimal.NewFromFloat32(0.2)).
				Mul(decimal.NewFromInt(100)).Float64()
			tmpScore += tmp
		}
		if sum.killUnit != 0 {
			tmp, _ := decimal.NewFromInt(item.killUnit).
				Div(decimal.NewFromInt(sum.killUnit)).
				Mul(decimal.NewFromFloat32(0.1)).
				Mul(decimal.NewFromInt(100)).Float64()
			tmpScore += tmp
		}
		if sum.deKillUnit != 0 {
			tmp, _ := decimal.NewFromInt(item.deKillUnit).
				Div(decimal.NewFromInt(sum.deKillUnit)).
				Mul(decimal.NewFromFloat32(0.1)).
				Mul(decimal.NewFromInt(100)).Float64()
			tmpScore += tmp
		}
		// z(0-100)=(xi-min(x))/(max(x)-min(x))*100

		itemResp.Score += float32(tmpScore)
		if item.firstBlood {
			itemResp.Score += 5
		}
		if item.deFirstBlood {
			itemResp.Score -= 5
		}
		if item.deKillPlayer {
			itemResp.Score -= 5
		}
		if item.isGeneral {
			itemResp.Score += 5
		}
		if itemResp.Score < 0 {
			itemResp.Score = 0
		}
		// 最终进行四舍五入
		tmp, _ := decimal.NewFromFloat32(itemResp.Score).Round(0).Float64()
		itemResp.Score = float32(tmp)

		item.respItem = itemResp
	}
}

func (l *StatPvpReportLogic) reports(win bool, items []*statPvPReportItem) error {
	return l.svcCtx.StatisticsPvpModel.TransactCtx(l.ctx, nil, func(tx *gorm.DB) error {
		for i, item := range items {
			item.respItem.Position = int32(i + 1)
			props := &model.UpdateRecordProps{
				Damage:          item.damage,
				DeDamage:        item.deDamage,
				KillUnitCount:   item.killUnit,
				DeKillUnitCount: item.deKillUnit,
				Win:             win,
				Lost:            !win,
				KillPlayer:      item.killPlayer,
				DeKillPlayer:    item.deKillPlayer,
				DeGeneral:       item.deKillPlayer && item.isGeneral,
				FirstBlood:      item.firstBlood,
				DeFirstBlood:    item.deFirstBlood,
			}
			if win && i == 0 {
				// 名将
				props.General = true
			}

			if err := l.svcCtx.StatisticsPvpModel.UpdateRecord(l.ctx, tx, item.uid, props); err != nil {
				if errors.Is(err, gormx.ErrRowsAffectedZero) {
					// insert
					var (
						winCount, lostCount, deKillPlayer, generalCount, firstBloodCount, deFirstBloodCount int64
					)
					if win {
						winCount = 1
					} else {
						lostCount = 1
					}
					if item.deKillPlayer {
						deKillPlayer = 1
					}
					if props.General {
						generalCount = 1
					}
					if props.FirstBlood {
						firstBloodCount = 1
					}
					if props.DeFirstBlood {
						deFirstBloodCount = 1
					}
					if err = l.svcCtx.StatisticsPvpModel.Insert(l.ctx, tx, &model.StatisticsPvp{
						UserId:            item.uid,
						WinCount:          winCount,
						LostCount:         lostCount,
						Damage:            item.damage,
						DeDamage:          item.deDamage,
						KillUnitCount:     item.killUnit,
						DeKillUnitCount:   item.deKillUnit,
						KillPlayerCount:   item.killPlayer,
						DeKillPlayerCount: deKillPlayer,
						GeneralCount:      generalCount,
						DeGeneralCount:    0,
						FirstBloodCount:   firstBloodCount,
						DeFirstBloodCount: deFirstBloodCount,
					}); err != nil {
						return errors.Wrapf(err, ErrInsertErr, err)
					}
				} else {
					return errors.Wrapf(err, ErrUpdateErr, err)
				}
			}
		}
		return nil
	})
}
