package user

import (
	"context"
	"git.noahlan.cn/northlan/ntools-go/uuid"
	"github.com/pkg/errors"
	"github.com/zeromicro/go-zero/core/logx"
	"gorm.io/gorm"
	pbVars "live-service/app/pb/vars"
	"live-service/app/user_center/model"
	"live-service/app/user_center/rpc/internal/common/coin_manager"
	"live-service/app/user_center/rpc/internal/svc"
	"live-service/app/user_center/rpc/pb"
	"live-service/common/nerr"
	"math/rand"
	"time"
)

type UserCheckInLogic struct {
	ctx    context.Context
	svcCtx *svc.ServiceContext
	logx.Logger
}

func NewUserCheckInLogic(ctx context.Context, svcCtx *svc.ServiceContext) *UserCheckInLogic {
	return &UserCheckInLogic{
		ctx:    ctx,
		svcCtx: svcCtx,
		Logger: logx.WithContext(ctx),
	}
}

const (
	MsgAlreadyCheckIn = "今天已打过卡了！"
)

// UserCheckIn 用户签到|打卡
func (l *UserCheckInLogic) UserCheckIn(in *pb.UserIdReq) (*pb.UserCheckInResp, error) {
	resp := &pb.UserCheckInResp{
		IsCritical: false,
	}
	if err := l.svcCtx.Db.Transaction(func(tx *gorm.DB) error {
		isCheckInToday, err := l.svcCtx.UserCheckInModel.CheckInToday(l.ctx, tx, in.UserId)
		if err != nil {
			return nerr.NewWithCode(nerr.DBError)
		}
		if isCheckInToday {
			return nerr.NewError(nerr.UserAlreadyCheckIn, MsgAlreadyCheckIn)
		}
		// 打卡记录
		if err = l.svcCtx.UserCheckInModel.Insert(l.ctx, tx, &model.UserCheckIn{
			Id:          uuid.NextId(),
			UserId:      in.UserId,
			CheckInTime: time.Now(),
		}); err != nil {
			return nerr.NewWithCode(nerr.DBError)
		}

		// 积分记录，计算当日可得积分
		change, critical := l.getCriticalPoint()

		changeCoinResp, err := l.svcCtx.CoinManager.ChangeCoin(l.ctx, tx, &coin_manager.ChangeCoinReq{
			UserId: in.UserId,
			Change: change,
		}, pbVars.UserCoinChangedReason_CheckIn)
		if err != nil {
			return errors.Wrapf(nerr.NewWithErr(err), "记录弹币失败, err:%+v", err)
		}
		resp.CurrentCoin = changeCoinResp.Current
		resp.CoinChange = change
		resp.IsCritical = critical
		return nil
	}); err != nil {
		return nil, errors.Wrapf(nerr.NewWithErr(err), "打卡失败 err:%v", err)
	}
	return resp, nil
}

func (l *UserCheckInLogic) getCriticalPoint() (int64, bool) {
	checkInCfg := l.svcCtx.GameConfig.Coin.CheckIn
	nt := time.Now()
	weekDay := int(nt.Weekday())
	point := checkInCfg.Points[weekDay]

	rand.Seed(nt.UnixNano())
	isCritical := rand.Float32() <= checkInCfg.Critical
	var radio float32
	if len(checkInCfg.CriticalRadio) == 0 {
		radio = 1
	} else {
		radio = checkInCfg.CriticalRadio[rand.Intn(len(checkInCfg.CriticalRadio))]
	}
	if isCritical {
		return int64(float32(point) * radio), isCritical
	} else {
		return point, isCritical
	}
}
